use crossterm::event::{Event as CEvent, KeyCode, KeyEvent};
use log::{debug, error, warn};
use specs::prelude::*;
use specs::saveload::{MarkedBuilder, UuidMarker};

use crate::model::menu::{
    RegisterItemStatus, RentItemStatus, ReturnItemStatus, Status as MenuStatus,
};
use crate::model::{Location, Metadata, State, TagId};

#[derive(Debug)]
pub enum Status {
    Exit,
    Continue,
}

pub fn on_event(state: &mut State, event: CEvent) -> Status {
    log::debug!("Recieved event: {:?}", event);

    match event {
        CEvent::Key(KeyEvent {
            code: KeyCode::Char('q'),
            ..
        }) if state.menu_status == MenuStatus::Main => Status::Exit,
        CEvent::Key(KeyEvent {
            code: KeyCode::Char(c),
            ..
        }) => {
            state.input.push(c);
            Status::Continue
        }
        CEvent::Key(KeyEvent {
            code: KeyCode::Backspace,
            ..
        }) => {
            state.input.pop();
            Status::Continue
        }
        CEvent::Key(KeyEvent {
            code: KeyCode::Enter,
            ..
        }) => {
            debug!("confirmed: {}", state.input);

            match state.menu_status {
                MenuStatus::Main => match &*state.input {
                    "1" => {
                        debug!("requested to register new item");
                        state.menu_status = MenuStatus::RegisterItem(
                            RegisterItemStatus::InsertItemId,
                        );
                    }
                    "2" => {
                        debug!("requested to rent item");
                        state.menu_status =
                            MenuStatus::RentItem(RentItemStatus::InsertRenter);
                    }
                    "3" => {
                        debug!("requested to return item");
                        state.menu_status = MenuStatus::ReturnItem(
                            ReturnItemStatus::InsertItemId,
                        );
                    }
                    _ => warn!("command not found"),
                },
                MenuStatus::RegisterItem(ref reg_status) => match reg_status {
                    RegisterItemStatus::InsertItemId => {
                        let id = state.input.clone();

                        if state
                            .world
                            .read_storage::<TagId>()
                            .join()
                            .find(|tag_id| tag_id.0 == id)
                            .is_none()
                        {
                            state.menu_status = MenuStatus::RegisterItem(
                                RegisterItemStatus::InsertItemName { id },
                            );
                        } else {
                            error!("item with id '{}' already exists", id);
                        }
                    }
                    RegisterItemStatus::InsertItemName { id } => {
                        let _item = state
                            .world
                            .create_entity()
                            .with(TagId(id.clone()))
                            .with(Location::InStorage)
                            .with(Metadata {
                                name: state.input.clone(),
                            })
                            .marked::<UuidMarker>()
                            .build();

                        state.menu_status = MenuStatus::Main;
                    }
                },
                MenuStatus::RentItem(ref rent_status) => match rent_status {
                    RentItemStatus::InsertRenter => {
                        let renter = state.input.clone();

                        state.menu_status = MenuStatus::RentItem(
                            RentItemStatus::InsertItemId { renter },
                        );
                    }
                    RentItemStatus::InsertItemId { renter } => {
                        let id = &state.input;

                        if let Some(item) = state.world.exec(
                            |(entities, tag_ids): (
                                Entities,
                                ReadStorage<TagId>,
                            )| {
                                (&entities, &tag_ids)
                                    .join()
                                    .find(|(_, tag_id)| &tag_id.0 == id)
                                    .map(|(e, _)| e)
                            },
                        ) {
                            if let Some(loc) = state
                                .world
                                .write_storage::<Location>()
                                .get_mut(item)
                            {
                                debug!(
                                    "item with id '{}' at {:?} rented to {}",
                                    id, loc, state.input
                                );
                                *loc = Location::RentedTo {
                                    renter: renter.clone(),
                                };
                            } else {
                                warn!("item with id '{}' does not have a location", id);
                            }
                        } else {
                            error!("there exists no item with id '{}'", id);
                        }
                    }
                },
                MenuStatus::ReturnItem(ref return_status) => {
                    match return_status {
                        ReturnItemStatus::InsertItemId => {
                            let id = &state.input;

                            if let Some(item) = state.world.exec(
                                |(entities, tag_ids, locs): (
                                    Entities,
                                    ReadStorage<TagId>,
                                    ReadStorage<Location>,
                                )| {
                                    (&entities, &tag_ids, &locs)
                                        .join()
                                        .find(|(_, tag_id, loc)| {
                                            &tag_id.0 == id
                                                && !matches!(
                                                    loc,
                                                    Location::InStorage
                                                )
                                        })
                                        .map(|(e, _, _)| e)
                                },
                            ) {
                                if let Some(loc) = state
                                    .world
                                    .write_storage::<Location>()
                                    .get_mut(item)
                                {
                                    debug!(
                                        "item '{}' returned from {:?}",
                                        id, loc
                                    );
                                    *loc = Location::InStorage;
                                } else {
                                    warn!(
                                        "item '{}' does not have a location",
                                        id
                                    );
                                }
                            } else {
                                error!(
                                "there exists no item with id '{}' which is not in storage",
                                id
                            );
                            }
                        }
                    }
                }
            }

            state.input.clear();
            Status::Continue
        }
        CEvent::Key(KeyEvent {
            code: KeyCode::Esc, ..
        }) => {
            state.menu_status = MenuStatus::Main;
            state.input.clear();
            Status::Continue
        }
        _ => Status::Continue,
    }
}

pub fn on_tick(_state: &mut State) {}
