use anyhow::Result;
use arraydeque::{behavior::Wrapping, ArrayDeque};
use log::*;
use std::sync::Arc;
use std::sync::Mutex;
use tui::buffer::Buffer;
use tui::layout::Rect;
use tui::style::{Color, Style};
use tui::widgets::Text;
use tui::widgets::Widget;

pub fn init(level: LevelFilter) -> Result<TuiLogState> {
    let logger = Box::new(TuiLogger {
        logs: Arc::new(Mutex::new(ArrayDeque::new())),
    });

    let log_ref = TuiLogState {
        logs: Arc::clone(&logger.logs),
    };

    set_boxed_logger(logger)?;
    set_max_level(level);

    Ok(log_ref)
}

#[derive(Debug)]
pub struct TuiLogState {
    logs: Arc<Mutex<ArrayDeque<[Text<'static>; 256], Wrapping>>>,
}

#[derive(Debug)]
pub struct TuiLogWidget<'s> {
    pub state: &'s TuiLogState,
}

impl Widget for TuiLogWidget<'_> {
    fn render(self, area: Rect, buf: &mut Buffer) {
        use tui::style::*;
        use tui::widgets::*;

        let logs = self.state.logs.lock().expect("logs lock posioned");

        let block = Block::default().title("Logs").borders(Borders::ALL);
        let p = Paragraph::new(logs.iter())
            .block(block)
            .style(Style::default().fg(Color::White).bg(Color::Black))
            .wrap(true);

        p.render(area, buf);
    }
}

#[derive(Debug)]
struct TuiLogger {
    // TODO maybe use something else than a mutex here??
    logs: Arc<Mutex<ArrayDeque<[Text<'static>; 256], Wrapping>>>,
}

impl Log for TuiLogger {
    fn enabled(&self, _metadata: &Metadata) -> bool {
        true
    }

    fn log(&self, record: &Record) {
        if !self.enabled(record.metadata()) {
            return;
        }

        let (label, color) = match record.level() {
            Level::Error => ("ERROR", Color::Red),
            Level::Warn => ("WARN", Color::Yellow),
            Level::Info => ("INFO", Color::Blue),
            Level::Debug => ("DEBUG", Color::Magenta),
            Level::Trace => ("TRACE", Color::Gray),
        };

        let mut logs = self.logs.lock().expect("logs lock posioned");
        logs.push_front(Text::raw(format!(
            " {}: {}\n",
            record.target(),
            record.args()
        )));
        logs.push_front(Text::Styled(label.into(), Style::new().fg(color)));
    }

    fn flush(&self) {}
}
