use anyhow::Result;
use futures::{future, select};
use futures::{FutureExt, StreamExt};
use log::{debug, error, info};
use specs::prelude::*;
use std::io::{self, Write};
use std::thread;
use std::time::Duration;
use tui::backend::CrosstermBackend;
use tui::Terminal;

mod logging;

mod controller;
mod model;
mod view;

use crate::controller::Status;

fn main() -> Result<()> {
    // Setup terminal for application
    crossterm::terminal::enable_raw_mode()?;
    let mut stdout = io::stdout();
    crossterm::execute!(stdout, crossterm::terminal::EnterAlternateScreen)?;
    let backend = CrosstermBackend::new(stdout);
    let mut terminal = Terminal::new(backend)?;
    terminal.hide_cursor()?;

    // Setup executors and work pool (required for using block_on later)
    for _ in 0..4 {
        thread::spawn(|| smol::run(future::pending::<()>()));
    }

    let log_state = logging::init(log::LevelFilter::Info)?;

    let mut state = model::State::new(log_state);

    {
        use crate::model::{Location, Metadata, TagId};
        use specs::prelude::*;
        use specs::saveload::{
            DeserializeComponents, UuidMarker, UuidMarkerAllocator,
        };
        use std::fs::File;

        debug!("trying to load state from 'data.json'");
        if let Ok(file) = File::open("data.json") {
            let mut de = serde_json::Deserializer::from_reader(file);

            let (ents, mut markers, mut alloc, tags, metas, locs): (
                Entities,
                WriteStorage<UuidMarker>,
                Write<UuidMarkerAllocator>,
                WriteStorage<TagId>,
                WriteStorage<Metadata>,
                WriteStorage<Location>,
            ) = state.world.system_data();

            match DeserializeComponents::<anyhow::Error, _>::deserialize(
                &mut (tags, metas, locs),
                &ents,
                &mut markers,
                &mut alloc,
                &mut de,
            ) {
                Ok(()) => info!("loaded state from 'data.json'"),
                Err(e) => error!(
                    "unable to read or parse data 'data.json' storage: {}",
                    e
                ),
            }
        }
    }

    // Generate an infinite stream of tick events to update UI (doesn't need to be fused because it
    // will never return `None`)
    let mut ticks = futures::stream::unfold((), |_| {
        // 16 ms per frame means roughly 60 FPS
        smol::Timer::after(Duration::from_millis(16))
            .map(|time| Some((time, ())))
    });

    // Get input events from crossterm (must be fused to be used in select)
    let mut terminal_events = crossterm::event::EventStream::new().fuse();

    // Run main application
    let app_result: Result<()> = smol::block_on(async {
        loop {
            select! {
                event = terminal_events.select_next_some() => {
                    match event {
                        Ok(event) => if let Status::Exit = controller::on_event(&mut state, event) { return Ok(()); },
                        Err(e) => log::error!("unable to get terminal event: {}", e),
                    }
                }
                tick = ticks.select_next_some() => {
                    state.world.maintain();
                    controller::on_tick(&mut state);
                    terminal.draw(|mut f| { view::draw(&state, f) })?;
                }
            }
        }
    });

    {
        use crate::model::{Location, Metadata, TagId};
        use specs::saveload::{SerializeComponents, UuidMarker};
        use std::fs::File;

        let mut file = File::create("data.json")?;
        let mut ser = serde_json::Serializer::new(&mut file);

        let (ents, markers, tags, metas, locs): (
            Entities,
            ReadStorage<UuidMarker>,
            ReadStorage<TagId>,
            ReadStorage<Metadata>,
            ReadStorage<Location>,
        ) = state.world.system_data();

        let _ = SerializeComponents::<anyhow::Error, UuidMarker>::serialize(
            &(&tags, &metas, &locs),
            &ents,
            &markers,
            &mut ser,
        );
    }

    // Reset terminal after application exits
    terminal.show_cursor()?;
    crossterm::execute!(
        terminal.backend_mut(),
        crossterm::terminal::LeaveAlternateScreen
    )?;
    crossterm::terminal::disable_raw_mode()?;

    app_result
}
